/*
 * FileVisitor.cs
 * 
 * Performs the actual search (and replace) operation on a given file.
 */
using System;
using System.IO;
using System.Text.RegularExpressions;

namespace Jitsu
{
    internal class FileVisitor
    {
        // For line- and column-num reporting
        internal static readonly Regex LineStartRTL = new Regex("^", RegexOptions.Compiled | RegexOptions.Multiline | RegexOptions.RightToLeft);
        internal static readonly Regex LineTermLTR = new Regex("$", RegexOptions.Compiled | RegexOptions.Multiline);

        internal int FilesSearched = 0, FilesFound = 0, MatchesFound = 0;

        // User arguments
        internal Regex UserRegex = null;
        internal string ReplacementText = null;

        // Config parameters
        internal ushort OutputWidth = 78;

        public void VisitFile(string relpath)
        {
            // Make a note of how we need to open files.
            FileAccess fileAccess = FileAccess.Read;
            FileShare fileShare = FileShare.Read;

            if (ReplacementText != null)
            {
                fileAccess = FileAccess.ReadWrite;
                fileShare = FileShare.None; ;
            }

            // Open the file, and read it in (default to ANSI encoding, but detecting BOM if present).
            try
            {
                using (FileStream fs = new FileStream(relpath, FileMode.Open, fileAccess, fileShare))
                using (StreamReader sr = new StreamReader(fs, System.Text.Encoding.Default, true))
                {
                    // Read the entire file into a string (inefficient, but req'd for regx).
                    string text = sr.ReadToEnd();

                    // Convert CRLF to LF -- this is doubly inefficient, but necessary to keep 
                    // RegexOptions.Multiline from matching the 0D char at the end of lines.
                    text = text.Replace("\r\n", "\n");

                    // Bump the hit count.
                    FilesSearched++;

                    // Start the search.
                    MatchCollection matches = UserRegex.Matches(text);
                    if (matches.Count > 0)
                    {
                        FilesFound++;
                        MatchesFound += matches.Count;

                        foreach (Match m in matches)
                        {
                            // Count number of line-starts to our left to get a 1-based row number.
                            int lineNum = LineStartRTL.Matches(text, m.Index).Count;

                            // Calculate the column number (1-based).
                            int columnNum = m.Index - LineStartRTL.Match(text, m.Index).Index + 1;

                            // Extend the match to include the whole line(s) of text, for display purposes.
                            int matchBegin = m.Index;
                            int matchEnd = m.Index + m.Length;

                            int lineBegin = LineStartRTL.Match(text, matchBegin).Index;
                            int lineEnd = LineTermLTR.Match(text, matchEnd).Index;

                            // For matches on very long lines: clip begin and end at +/- 40 chars from match.
                            int lineWidth = lineEnd - lineBegin;
                            if (lineWidth > OutputWidth)
                            {
                                int lineBegin0 = lineBegin;
                                int lineEnd0 = lineEnd;
                                int middle = m.Index + m.Length / 2;

                                lineBegin = Math.Min(Math.Max(lineBegin0, middle - OutputWidth / 2), matchBegin);
                                lineEnd = Math.Min(Math.Max(matchEnd, middle + OutputWidth / 2), lineEnd0);

                                // But if we're up against one side or the other, extend outward in the other direction.
                                if (lineEnd - lineBegin < OutputWidth)
                                {
                                    lineBegin = Math.Max(lineBegin0, lineEnd - OutputWidth);
                                    lineEnd = Math.Min(lineEnd0, lineBegin + OutputWidth);
                                }
                            }

                            // Populate 3 strings for output in different colors: head, match, tail
                            string lineTextHead = text.Substring(lineBegin, matchBegin - lineBegin);

                            string lineTextMatch = text.Substring(matchBegin, m.Length);

                            string lineTextTail = text.Substring(matchEnd, lineEnd - matchEnd);

                            // Replace tabs with single-spaces.
                            lineTextHead = lineTextHead.Replace('\t', ' ');
                            lineTextMatch = lineTextMatch.Replace('\t', ' ');
                            lineTextTail = lineTextTail.Replace('\t', ' ');

                            // Replace linefeeds with  markers.
                            lineTextHead = lineTextHead.Replace('\n', '');
                            lineTextMatch = lineTextMatch.Replace('\n', '');
                            lineTextTail = lineTextTail.Replace('\n', '');

                            // Print the file name and match location.
                            Console.WriteLine();
                            SafeConsoleColorSetter.SetColor(ConsoleColor.Green);
                            Console.Write("{0}", Path.GetFullPath(relpath));
                            SafeConsoleColorSetter.SetColor(ConsoleColor.White);
                            Console.Write("({0},{1}): ", lineNum, columnNum);

                            SafeConsoleColorSetter.ResetColor();
                            Console.WriteLine();

                            // Print the match, in 3 syntax-highlighted parts.
                            SafeConsoleColorSetter.ResetColor();
                            Console.Write("{0}", lineTextHead);
                            SafeConsoleColorSetter.SetColor(ConsoleColor.Gray, ConsoleColor.DarkGray);
                            Console.Write("{0}", lineTextMatch);
                            SafeConsoleColorSetter.ResetColor();
                            SafeConsoleColorSetter.ResetColor();
                            Console.Write("{0}", lineTextTail);

                            SafeConsoleColorSetter.ResetColor();
                            Console.WriteLine();
                        }

                        if (ReplacementText != null)
                        {
                            // Reset the stream-length and file-pointer -- write the modified text, taking care to preserve 
                            // the original encoding.
                            fs.SetLength(0);
                            using (StreamWriter sw = new StreamWriter(fs, sr.CurrentEncoding))
                            {
                                // Slam it down.
                                text = UserRegex.Replace(text, ReplacementText);

                                // Restore platform line-endings, then save.
                                text = text.Replace("\n", Environment.NewLine);
                                sw.Write(text);
                            }
                        }
                    }
                }
            }
            catch (IOException ex)
            {
                SafeConsoleColorSetter.SetColor(ConsoleColor.Yellow);
                Console.Error.WriteLine(ex.Message);
                SafeConsoleColorSetter.ResetColor();
            }
        }
    }
}